/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.
*/

#include "chirp_protocol.h"
#include "chirp_acl.h"
#include "chirp_exec.h"
#include "chirp_client.h"

#include "macros.h"
#include "debug.h"
#include "link.h"
#include "getopt.h"
#include "auth_all.h"
#include "stringtools.h"
#include "full_io.h"
#include "get_canonical_path.h"
#include "datagram.h"
#include "username.h"
#include "catalog_server.h"
#include "disk_info.h"
#include "domain_name_cache.h"
#include "create_dir.h"
#include "delete_dir.h"
#include "list.h"
#include "xmalloc.h"
#include "md5.h"
#include "limit.h"
#include "quota.h"
#include "load_average.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <limits.h>
#include <sys/stat.h>
#ifdef HAS_SYS_STATFS_H
#include <sys/statfs.h>
#endif
#ifdef HAS_SYS_STATVFS_H
#include <sys/statvfs.h>
#endif
#include <sys/mount.h>
#include <sys/param.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <signal.h>
#include <utime.h>
#include <sys/wait.h>
#include <pwd.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/utsname.h>

/* Cygwin does not have 64-bit I/O, while Darwin has it by default. */

#if CCTOOLS_OPSYS_CYGWIN || CCTOOLS_OPSYS_DARWIN
#define fopen64 fopen
#define open64 open
#define lseek64 lseek
#define stat64 stat
#define fstat64 fstat
#define lstat64 lstat
#define fseeko64 fseeko
#endif

/* Solaris has statfs, but it doesn't work! Use statvfs instead. */
#if CCTOOLS_OPSYS_SUNOS
#define statfs statvfs
#define fstatfs fstatvfs
#endif

#if CCTOOLS_OPSYS_DARWIN
typedef int blkcnt_t;
typedef int blksize_t;
#endif
 
/* The maximum chunk of memory the server will allocate to handle I/O */
#define MAX_BUFFER_SIZE (16*1024*1024)

static void chirp_receive( struct link *l );
static void chirp_handler( struct link *l, const char *subject );
static int errno_to_chirp( int e );

static int port = CHIRP_PORT;
static int idle_timeout = 86400; /* one day */
static int stall_timeout = 3600; /* one hour */
static int advertise_timeout = 300; /* five minutes */
static int advertise_alarm = 0;
static int single_mode = 0;
static struct list *catalog_host_list;
static char owner[USERNAME_MAX] = "";
static time_t starttime;
static struct datagram *catalog_port;
static char hostname[DOMAIN_NAME_MAX];
static const char *manual_hostname = 0;
static char address[LINK_ADDRESS_MAX];
static char rootpath[CHIRP_PATH_MAX];
static const char *startdir = ".";
static const char *safe_username = 0;
static uid_t safe_uid = 0;
static gid_t safe_gid = 0;
static int allow_execute = 0;
static int did_acl_default = 0;
static int dont_dump_core = 0;
int enable_identity_boxing = 1;
unsigned long int bytes_per_second = 0;
unsigned int bandwidth_period_ms = 200;
INT64_T bandwidth_refill = 0;
int use_quotas = 0;
struct limit bandwidth_limit;
blksize_t max_disk_blocks;
const char *chirp_server_path = 0;

struct server_stats {
	INT64_T bytes_read;
	INT64_T bytes_written;
	INT64_T files_opened;
	INT64_T total_clients;
        INT64_T d_rbytes_max;
        INT64_T d_wbytes_max;
};

struct server_stats *global_stats = 0;

static void show_version( const char *cmd )
{
        printf("%s version %d.%d.%d built by %s@%s on %s at %s\n",cmd,CCTOOLS_VERSION_MAJOR,CCTOOLS_VERSION_MINOR,CCTOOLS_VERSION_MICRO,BUILD_USER,BUILD_HOST,__DATE__,__TIME__);
}

static void show_help( const char *cmd )
{
	printf("use: %s [options]\n",cmd);
	printf("where options are:\n");
	printf(" -d <flag>   Enable debugging for this sybsystem\n");
	printf(" -c <dir>    Challenge directory for filesystem authentication.\n");
	printf(" -C          Do not create a core dump, even due to a crash.\n");
	printf(" -h          This message.\n");
	printf(" -o <file>   Send debugging output to this file.\n");
	printf(" -O <bytes>  Rotate debug file once it reaches this size.\n");
	printf(" -n <name>   Use this name when reporting to the catalog.\n");
	printf(" -p <port>   Listen on this port (default is %d)\n",port);
	printf(" -A <file>   Use this file as the default ACL.\n");
	printf(" -r <dir>    Root of storage directory. (default is current dir)\n");
	printf(" -R          Read-only / read-everything mode.\n");
	printf(" -s <time>   Abort stalled operations after this long. (default is 1h)\n");
	printf(" -S          Single process mode, do not fork.\n");
	printf(" -t <time>   Disconnect idle clients after this time. (default is 1d)\n");
	printf(" -u <host>   Send status updates to this host. (default is %s)\n",CATALOG_HOST);
	printf(" -U <time>   Send status updates at this interval. (default is 5m)\n");
	printf(" -v          Show version info.\n");
	printf(" -w <name>   The name of this server's owner.  (default is username)\n");
	printf(" -X          Enable remote execution.\n");
	printf(" -N          Disable identity boxing for execution.  (discouraged)\n");
	printf(" -B <B/s>    Limit bandwidth consumption (default unlimited) (experimental)\n");
	printf(" -P <ms>     Period for bandwidth checks (default %u ms) (experimental)\n", bandwidth_period_ms );
        printf(" -q          Use user-level quotas (experimental)\n");
	printf(" -Q <KB>     Use this amount as default root quota (experimental)\n");
	printf("\n");
	printf("Where debug flags are: ");
	debug_flags_print(stdout);
	printf("\n\n");
	printf("Time arguments to -t, -s, and -U are assumed to be in seconds unless followed\n");
	printf("by an 's' 'm' 'h' or 'd' to indicate seconds, minutes, hours, or days.\n");
	printf("For example: chirp_server -t 7d -s 2h\n");
}

void shutdown_clean( int sig )
{
	exit(0);
}

void ignore_signal( int sig )
{
}

void reap_child( int sig )
{
	pid_t pid;
	int status;

	do {
		pid = waitpid(-1,&status,WNOHANG);
	} while(pid>0);
}

void refill_bandwidth( int sig ) {
	global_stats->d_rbytes_max = bandwidth_refill;
	global_stats->d_wbytes_max = bandwidth_refill;
	kill( 0, SIGCONT );
}

void wakeup( int sig ) {
}

static void install_handler( int sig, void (*handler)(int sig))
{
	struct sigaction s;
	s.sa_handler = handler;
	sigfillset(&s.sa_mask);
	s.sa_flags = 0; 
	sigaction(sig,&s,0);
}

int update_one_catalog( void *catalog_host, const void *text )
{
	char addr[DATAGRAM_ADDRESS_MAX];
	if(domain_name_cache_lookup(catalog_host,addr)) {
		debug(D_DEBUG,"sending update to %s:%d",catalog_host,CATALOG_PORT);
		datagram_send(catalog_port,text,strlen(text),addr,CATALOG_PORT);
	}
	return 1;
}

static void update_all_catalogs()
{
	struct utsname name;
	char text[DATAGRAM_PAYLOAD_MAX];
	UINT64_T avail, total;
	unsigned uptime;
	int length;
	double avg[3];

	uname(&name);
	string_tolower(name.sysname);
	string_tolower(name.machine);
	string_tolower(name.release);
	load_average_get(avg);
	disk_info_get(".",&avail,&total);
	uptime = time(0)-starttime;

	length = sprintf(
		text,
		"type chirp\nversion %d.%d.%d\nurl chirp://%s:%d\nname %s\nowner %s\ntotal %llu\navail %llu\nuptime %u\nport %d\nbytes_written %lld\nbytes_read %lld\nfiles_opened %lld\ntotal_clients %lld\ncpu %s\nopsys %s\nopsysversion %s\nbandwidth %ld\nload1 %0.02lf\nload5 %0.02lf\nload15 %0.02lf\n",		
		CCTOOLS_VERSION_MAJOR,
		CCTOOLS_VERSION_MINOR,
		CCTOOLS_VERSION_MICRO,
		hostname,
		port,
		hostname,
		owner,
		total,
		avail,
		uptime,
		port,
		global_stats->bytes_written,
		global_stats->bytes_read,
		global_stats->files_opened,
		global_stats->total_clients,
		name.machine,
		name.sysname,
		name.release,
		bytes_per_second,
		avg[0],
		avg[1],
		avg[2]
		);

	list_iterate(catalog_host_list,update_one_catalog,text);
}

int main( int argc, char *argv[] )
{
	struct link *link;
	char c;
	int c_input;
	struct itimerval itval;
	struct timeval  tval;
	time_t current;
	long inputval;

	chirp_server_path = argv[0];

	catalog_host_list = list_create();

	debug_config(argv[0]);

	while((c_input = getopt( argc,argv,"A:a:c:Ct:i:s:Sn:p:r:Ro:O:d:vw:u:U:hXNB:P:qQ:D")) != -1 ) {
	        c = ( char ) c_input;
		switch(c) {
		case 'A':
			chirp_acl_default(optarg);
			did_acl_default = 1;
			break;
		case 'a':
			fprintf(stderr,"The -a authfile option is no longer used.\n");
			fprintf(stderr,"You should be using Chirp ACLs instead.\n");
			fprintf(stderr,"Please the manual for more details.\n");
			exit(1);
			break;
		case 'c':
			auth_unix_challenge_dir(optarg);
			break;
		case 'C':
			dont_dump_core = 1;
			break;
		case 'd':
			debug_flags_set(optarg);
			break;
		case 'i':
			safe_username = optarg;
			break;
		case 'n':
			manual_hostname = optarg;
			break;			
		case 'p':
			port = atoi(optarg);
			break;
		case 't':
			idle_timeout = string_time_parse(optarg);
			break;
		case 's':
			stall_timeout = string_time_parse(optarg);
			break;
		case 'S':
			single_mode = 1;
			break;
		case 'r':
			startdir = optarg;
			break;
		case 'R':
			chirp_acl_force_readonly();
			break;
		case 'o':
			debug_config_file(optarg);
			break;
		case 'O':
			debug_config_file_size(string_metric_parse(optarg));
			break;
		case 'u':
			list_push_head(catalog_host_list,xstrdup(optarg));
			break;
		case 'U':
			advertise_timeout = string_time_parse(optarg);
			break;
		case 'v':
			show_version(argv[0]);
			return 1;
		case 'w':
			strcpy(owner,optarg);
			break;
		case 'X':
			allow_execute = 1;
			break;
		case 'N':
			enable_identity_boxing = 0;
			break;
		case 'B':
			inputval = atol( optarg );
			if( inputval < 0 ) {
				printf("Error: bytes per second cannot be "
				       "negative\n");
				show_help(argv[0]);
				return 1;
			}
			else
				bytes_per_second = inputval;
			break;
		case 'P':
			inputval = atol( optarg );
			if( inputval < 0 ) {
				printf("Error: seconds cannot be negative\n");
				show_help(argv[0]);
				return 1;
			}
			else
				bandwidth_period_ms = inputval;
			break;
		case 'D':
			//dmalloc_file = optarg;
			break;
		case 'q':
			use_quotas = 1;
			break;
		case 'Q':
			inputval = atol( optarg );
			if( inputval == 0 )
				chirp_acl_force_readonly();
			else if( inputval < 0 ){
				printf("Error: disk blocks cannot be negative\n");
				show_help(argv[0]);
				return 1;
			}
			else {
				//since VFS uses 512 byte blocks, multiply
				//KB by 2 to get VFS blocks
				max_disk_blocks = ( blksize_t ) inputval * 2;
			}
			break;
		case 'h':
		default:
			show_help(argv[0]);
			return 1;
		}
	}

	if(dont_dump_core) {
		struct rlimit rl;
		rl.rlim_cur = rl.rlim_max = 0;
		setrlimit(RLIMIT_CORE,&rl);
	}

	current = time(0);
	debug(D_ALL,"*** %s starting at %s",argv[0],ctime(&current));

	if(!list_size(catalog_host_list)) {
		list_push_head(catalog_host_list,CATALOG_HOST);
	}

	if(getuid()==0) {
		struct passwd *p;
		if(!safe_username) {		
			printf("Sorry, I refuse to run as root without certain safeguards.\n");
			printf("Please give me a safe username with the -i <user> option.\n");
			printf("After using root access to authenticate users,\n");
			printf("I will use the safe username to access data on disk.\n");
			exit(1);
		}
		p = getpwnam(safe_username);
		if(!p) fatal("unknown user: %s",safe_username);
		safe_uid = p->pw_uid;
		safe_gid = p->pw_gid;
		chown(startdir,safe_uid,safe_gid);
		chmod(startdir,0700);
	} else {
		if(safe_username) {
			printf("Sorry, the -i option doesn't make sense\n");
			printf("unless I am already running as root.\n");
			exit(1);
		}
	}

	if(!create_dir(startdir,0700))
		fatal("couldn't create %s: %s\n",startdir,strerror(errno));

	/* It's ok if this fails because there is a default permission check. */
        /* Note that it might fail if we are exporting a read-only volume. */
        chirp_acl_init_root(startdir);

	if(chdir(startdir)!=0)
		fatal("couldn't move to %s: %s\n",startdir,strerror(errno));
	if(getcwd(rootpath,sizeof(rootpath))<0)
		fatal("couldn't get working dir: %s\n",strerror(errno));

	link = link_serve(port);
	if(!link) fatal("couldn't listen on port %d: %s",port,strerror(errno));

	link_address_local(link,address,&port);

	auth_register_all();
	starttime = time(0);
	catalog_port = datagram_create(0);
	if(manual_hostname) {
		strcpy(hostname,manual_hostname);
	} else {
		domain_name_cache_guess(hostname);
	}
	if(!owner[0]) {
		if(!username_get(owner)) {
			strcpy(owner,"unknown");
		}
	}

	install_handler(SIGPIPE,ignore_signal);
	install_handler(SIGHUP,ignore_signal);
	install_handler(SIGCHLD,reap_child);
	install_handler(SIGINT,shutdown_clean);
	install_handler(SIGTERM,shutdown_clean);
	install_handler(SIGQUIT,shutdown_clean);
	install_handler(SIGXFSZ,ignore_signal);


#ifndef MAP_ANONYMOUS
#define MAP_ANONYMOUS MAP_ANON
#endif

	global_stats = mmap(0,sizeof(*global_stats),PROT_READ|PROT_WRITE,MAP_SHARED|MAP_ANONYMOUS,-1,0);
	if(global_stats==(void*)-1) fatal("couldn't create shared memory area: %s\n",strerror(errno));

	memset(global_stats,0,sizeof(*global_stats));

	//set up bandwidth management, if need be
	if( bytes_per_second > 0 ) {
		//calculate d_bytes_max
		bandwidth_refill = bytes_per_second * bandwidth_period_ms / 1000;

		//fill the bandwidth
		global_stats->d_rbytes_max = bandwidth_refill;
		global_stats->d_wbytes_max = bandwidth_refill;

		//install the handler
		install_handler( SIGALRM, refill_bandwidth );

		//set the timer to automatically reschedule
		tval.tv_usec = bandwidth_period_ms * 1000;
		tval.tv_sec  = 0;
		itval.it_interval = tval;
		itval.it_value    = tval;
		setitimer( ITIMER_REAL, &itval, NULL );

		debug( D_CHIRP, "bandwidth restriction enabled, period: %ld ms, "
		       "max transfer: %ld B/period", bandwidth_period_ms,
		       bandwidth_refill );
	}

	//set up user-level quotas here

	while(1) {
		struct link *l;
		pid_t pid;

		if(time(0)>advertise_alarm) {
			update_all_catalogs();
			advertise_alarm = time(0)+advertise_timeout;
		}

		l = link_accept(link,time(0)+advertise_timeout);
		if(!l) continue;

		if( bytes_per_second > 0 )
			//enable link quotas
			link_quota( l, &( global_stats->d_rbytes_max ),
				    &( global_stats->d_wbytes_max ) );

		if(single_mode) {
			chirp_receive(l);
		} else {
			pid = fork();
			if(pid==0) {
				/* turn off the automatic waitpid for
				   grandchildren */
				install_handler(SIGCHLD,ignore_signal);

				//disable the refill bandwidth alarm, if any
				alarm(0);

				//set up SIGCONT signal handler so
				//links wake up
				install_handler( SIGCONT, wakeup );

				//serve the connection
				chirp_receive(l);

				//bye
				_exit(0);
			} else if(pid>=0) {
				debug(D_CHIRP,"created pid %d",pid);
			} else {
				debug(D_CHIRP,"couldn't fork: %s",strerror(errno));
			}
			link_close(l);
		}
	}
}

static void chirp_receive( struct link *link )
{
	char *atype, *asubject;
	char typesubject[AUTH_TYPE_MAX+AUTH_SUBJECT_MAX];
	char addr[LINK_ADDRESS_MAX];
	int port;

	link_address_remote(link,addr,&port);
	debug(D_AUTH,"client connected from %s:%d",addr,port);

	if(auth_accept(link,&atype,&asubject,time(0)+stall_timeout)) {
		sprintf(typesubject,"%s:%s",atype,asubject);
		free(atype);
		free(asubject);
		debug(D_AUTH,"client authenticates as %s",typesubject);
		if(safe_username) {
			debug(D_AUTH,"changing to uid %d gid %d",safe_uid,safe_gid);
			setgid(safe_gid);
			setuid(safe_uid);
		}			
		global_stats->total_clients++;
		chirp_handler(link,typesubject);
		global_stats->total_clients--;
	} else {
		debug(D_AUTH,"client could not authenticate");
	}

	link_close(link);
}

/*
  Force a path to fall within the simulated root directory.
*/

static int chirp_path_fix( char *path )
{
	char temp[CHIRP_PATH_MAX*2];
	int rlen;

	// Put the root path into place:
	rlen = sprintf(temp,"%s/",rootpath);

	// Now append the canonicalized path with dots and such removed.
	string_collapse_path(path,&temp[rlen],1);

	// Copy the fixed path back out.
	strcpy(path,temp);

	return 1;
}

/*
  A note on integers:
  Various operating systems employ integers of different sizes
  for fields such as file size, user identity, and so forth.
  Regardless of the operating system support, the Chirp protocol
  must support integers up to 64 bits.  So, in the server handling
  loop, we treat all integers as INT64_T.  What the operating system
  does from there is out of our hands.
*/

static void chirp_handler( struct link *l, const char *subject )
{
	char line[CHIRP_LINE_MAX];
	
	link_tune(l,LINK_TUNE_INTERACTIVE);

	while(1) {
		int do_stat_result = 0;
		int do_statfs_result = 0;
		int do_getdir_result = 0;
		int do_no_result = 0;
		INT64_T result = -1;

		char *dataout = 0;
		INT64_T dataoutlength = 0;

		char path[CHIRP_PATH_MAX];
		char newpath[CHIRP_PATH_MAX];
		char newsubject[CHIRP_LINE_MAX];
		char newacl[CHIRP_LINE_MAX];
		char hostname[CHIRP_LINE_MAX];

		char infile[CHIRP_PATH_MAX];
		char outfile[CHIRP_PATH_MAX];
		char errfile[CHIRP_PATH_MAX];
		
		INT64_T fd, length, flags, offset, actual;
		INT64_T uid, gid, mode;
		struct stat64 statbuf;
		struct statfs statfsbuf;
		struct utimbuf ut;
		time_t idletime = time(0)+idle_timeout;
		time_t stalltime;
		char args[CHIRP_LINE_MAX];

		if(!link_readline(l,line,sizeof(line),idletime)) {
			if(time(0)>=idletime) {
				debug(D_CHIRP,"timeout: client idle too long\n");
				sprintf(line,"%d\n",CHIRP_ERROR_TIMED_OUT);
				link_write(l,line,strlen(line),time(0)+stall_timeout);
			}
			break;
		}

		stalltime = time(0)+stall_timeout;
		string_chomp(line);
		if(strlen(line)<1) continue;
		if(line[0]==4) break;

		debug(D_CHIRP,"%s",line);

		if(sscanf(line,"pread %lld %lld %lld",&fd,&length,&offset)==3) {
			length = MIN(length,MAX_BUFFER_SIZE);
			dataout = malloc(length);
			if(dataout) {
				result = full_pread64(fd,dataout,length,offset);
				if(result>=0) {
					dataoutlength = result;
					global_stats->bytes_read += result;
				} else {
					free(dataout);
					dataout = 0;
				}
			} else {
				errno = ENOMEM;
			}
		} else if(sscanf(line,"whoami %lld",&length)==1) {
			if(strlen(subject)<length) length = strlen(subject);
			dataout = malloc(length);
			if(dataout) {
				dataoutlength = length;
				strncpy(dataout,subject,length);
				result = length;
			} else {
				result = -1;
			}
		} else if(sscanf(line,"whoareyou %s %lld",hostname,&length)==2) {
			struct chirp_client *client;
			char hostsubject[CHIRP_LINE_MAX];
			int hostsize;
		  
			client = chirp_client_connect( hostname, 1, time(0) +
						       stall_timeout );
			if( client ) {
				hostsize = chirp_client_whoami( client, hostsubject,
								CHIRP_LINE_MAX,
								time(0) +
								stall_timeout );
				if( hostsize < length )
					length = hostsize;
				dataout = malloc( length );
				if( dataout ){
					dataoutlength = length;
					strncpy( dataout, hostsubject, length );
					result = hostsize;
				}
				else
					result = -1;
				chirp_client_disconnect( client );
			}
			else
				result = -1;
		  
		} else if(sscanf(line,"readlink %s %lld",path,&length)==2) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_READ)) goto failure;
			dataout = malloc(length);
			if(dataout) {
				result = readlink(path,dataout,length);
				if(result>=0) {
					/* Careful: readlink could expose the local root path. */
					/* If it matches, remove the prefix, otherwise panic. */
					int rlen = strlen(rootpath);
					if(!strncmp(rootpath,dataout,rlen)) {
						strcpy(dataout,&dataout[rlen]);
						result = result-rlen;
						dataoutlength = result;
					} else {
						free(dataout);
						dataout = 0;
						errno = EINVAL;
						result = -1;
					}
				} else {
					free(dataout);
					dataout = 0;
				}
			} else {
				errno = ENOMEM;
				result = -1;
			}
		} else if(sscanf(line,"pwrite %lld %lld %lld",&fd,&length,&offset)==3) {
			char *data;
			INT64_T orig_length = length;
			/* Don't accept really enormous writes */
			length = MIN(length,MAX_BUFFER_SIZE);
			data = malloc(length);
			if(data) {
				actual = link_read(l,data,length,stalltime);
				if(actual!=length) break;
				/* read the amount of data requested */
				result = full_pwrite(fd,data,length,offset);
				/* remove the amount of data declined */
				link_soak(l,(orig_length-result),stalltime);
				free(data);
				if(result>0) global_stats->bytes_written += result;
			} else {
				debug(D_CHIRP, "malloc failed; soaking %lld bytes",orig_length);
				link_soak(l,orig_length,stalltime);
				result = -1;
				errno = ENOMEM;
				break;
			}
		} else if(sscanf(line,"getdir %s",path)==1) {
			DIR *dir;
			struct dirent *d;

			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check_dir(path,subject,CHIRP_ACL_LIST)) goto failure;

			dir = opendir(path);
			if(dir) {
				link_write(l,"0\n",2,stalltime);
				while((d=readdir(dir))) {
					if(!strcmp(d->d_name,CHIRP_ACL_BASE_NAME)) continue;
					sprintf(line,"%s\n",d->d_name);
					link_write(l,line,strlen(line),stalltime);
				}
				closedir(dir);
				do_getdir_result = 1;
				result = 0;
			} else {
				result = -1;
			}
		} else if(sscanf(line,"getacl %s",path)==1) {
			char aclsubject[CHIRP_LINE_MAX];
			int aclflags;
			FILE *aclfile;

			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check_dir(path,subject,CHIRP_ACL_LIST)) goto failure;

			aclfile = chirp_acl_open(path);
			if(aclfile) {
				link_write(l,"0\n",2,stalltime);
				while(chirp_acl_read(aclfile,aclsubject,&aclflags)) {
					sprintf(line,"%s %s\n",aclsubject,chirp_acl_flags_to_text(aclflags));
					link_write(l,line,strlen(line),stalltime);
				}
				chirp_acl_close(aclfile);
				do_getdir_result = 1;
				result = 0;
			} else {
				result = -1;
			}
		} else if(sscanf(line,"getfile %s",path)) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_READ)) goto failure;

			fd = open64(path,O_RDONLY);
			if(fd>=0) {
				fstat64(fd,&statbuf);
				if(S_ISDIR(statbuf.st_mode)) {
					close(fd);
					result = -1;
					errno = EISDIR;
				} else {
					length = statbuf.st_size;
					sprintf(line,"%lld\n",length);
					link_write(l,line,strlen(line),stalltime);
					result = link_stream_from_fd(l,fd,length,stalltime);
					close(fd);
					if(result!=length) break;
					do_no_result = 1;
					global_stats->files_opened++;
					global_stats->bytes_read += length;
				}
			} else {
				result = -1;
			}
		} else if(sscanf(line,"putfile %s %lld %lld",path,&mode,&length)) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_WRITE)) goto failure;

			/* remove the file to overcome permissions issues */
			unlink(path);

			fd = open64(path,O_WRONLY|O_CREAT|O_TRUNC,mode);
			if(fd>=0) {
				sprintf(line,"0\n");
				link_write(l,line,strlen(line),stalltime);
				actual = link_stream_to_fd(l,fd,length,stalltime);
				if(actual!=length) {
					if(actual>=0) link_soak(l,length-actual,stalltime);
					result = -1;
				} else {
					result = actual;
					global_stats->files_opened++;
					global_stats->bytes_written += length;
				}
				close(fd);
			} else {
				result = -1;
			}
		} else if(sscanf(line,"thirdput %s %s %s",path,hostname,
				 newpath)) {
			int save_errno;
			FILE *stream;
			struct chirp_client *client;
			struct stat64 info;
		  
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_READ))
				goto failure;

			//get file info/check for existence
			if( lstat64( path, &info ) >= 0 ) {
				//this command only works for regular files right now
				if( S_ISREG(info.st_mode) ) {
					stream = fopen64( path, "r" );

					if( stream ) {

						//authenticate with remote server
						client = chirp_client_connect( hostname, 1, time(0) +
									       stall_timeout );

						//connection successful?
						if( client ) {
							//issue putfile command
							length = chirp_client_putfile( client, newpath,
										       stream, info.st_mode,
										       info.st_size,
										       time(0) +
										       stall_timeout );

							//determine results
							if(length >=0 ) {
								fclose( stream );
								result = 0;
							}
							else {
								save_errno = errno;
								fclose( stream );
								errno = save_errno;
								result = -1;
							}

							//close connection
							chirp_client_disconnect( client );
						}
						else {
							//could not connect to third party
							save_errno = errno;
							result = -1;
							fclose( stream );
							errno = save_errno;
						}
					}
					else {
						//could not open file
						result = -1;
					}
				}
				else {
					//file is wrong type
					result = -1;
				}
			}
			else {
				//lstat failed
				result = -1;
			}

		} else if(sscanf(line,"open %s %s %lld",path,newpath,&mode)==3) {
			flags = 0;

			if(strchr(newpath,'r')) {
				if(strchr(newpath,'w')) {
					flags = O_RDWR;
				} else {
					flags = O_RDONLY;
				}
			} else if(strchr(newpath,'w')) {
				flags = O_WRONLY;
			}

			if(strchr(newpath,'c')) flags |= O_CREAT;
			if(strchr(newpath,'t')) flags |= O_TRUNC;
			if(strchr(newpath,'a'))	flags |= O_APPEND;
			if(strchr(newpath,'x')) flags |= O_EXCL;
#ifdef O_SYNC
			if(strchr(newpath,'s')) flags |= O_SYNC;
#endif

			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,chirp_acl_from_open_flags(flags))) goto failure;

			result = open64(path,flags,mode);
			if(result>=0) {
				fstat64(result,&statbuf);
				if(S_ISDIR(statbuf.st_mode)) {
					close(result);
					result = -1;
					errno = EISDIR;
				} else {
					global_stats->files_opened++;
					do_stat_result = 1;
				}
			}

		} else if(sscanf(line,"close %lld",&fd)==1) {
			result = close(fd);
		} else if(sscanf(line,"fchmod %lld %lld",&fd,&mode)==2) {
			result = fchmod(fd,mode);			
		} else if(sscanf(line,"fchown %lld %lld %lld",&fd,&uid,&gid)==3) {
			result = 0;
		} else if(sscanf(line,"fsync %lld",&fd)==1) {
			result = fsync(fd);
		} else if(sscanf(line,"ftruncate %lld %lld",&fd,&length)==2) {
			result = ftruncate(fd,length);
		} else if(sscanf(line,"unlink %s",path)==1) {
			if(!chirp_path_fix(path)) goto failure;
			if(
				chirp_acl_check(path,subject,CHIRP_ACL_WRITE) ||
				chirp_acl_check(path,subject,CHIRP_ACL_DELETE) ||
				chirp_acl_check_dir(path,subject,CHIRP_ACL_WRITE) ||
				chirp_acl_check_dir(path,subject,CHIRP_ACL_DELETE)
				) {
				result = unlink(path);
			} else {
				goto failure;
			}
		} else if(sscanf(line,"access %s %lld",path,&flags)==2) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,chirp_acl_from_access_flags(flags))) goto failure;
			result = access(path,flags);
		} else if(sscanf(line,"chmod %s %lld",path,&mode)==2) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_WRITE)) goto failure;
			result = chmod(path,mode);
		} else if(sscanf(line,"chown %s %lld %lld",path,&uid,&gid)==3) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_WRITE)) goto failure;
			result = 0;
		} else if(sscanf(line,"lchown %s %lld %lld",path,&uid,&gid)==3) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_WRITE)) goto failure;
			result = 0;
		} else if(sscanf(line,"truncate %s %lld",path,&length)==2) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_WRITE)) goto failure;
			result = truncate(path,length);
		} else if(sscanf(line,"rename %s %s",path,newpath)==2) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_path_fix(newpath)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_READ|CHIRP_ACL_DELETE) &&
			   !chirp_acl_check(path,subject,CHIRP_ACL_READ|CHIRP_ACL_WRITE) )
				goto failure;
			if(!chirp_acl_check(newpath,subject,CHIRP_ACL_READ|CHIRP_ACL_WRITE))
				goto failure;
			result = rename(path,newpath);
		} else if(sscanf(line,"link %s %s",path,newpath)==2) {
			/*
			  temporarily disabling links beause of a big security hole!
			  if(!chirp_path_fix(path)) goto failure;
			  if(!chirp_path_fix(newpath)) goto failure;
			  if(!chirp_acl_check(newpath,subject,CHIRP_ACL_WRITE)) goto failure;
		  
			  result = link(path,newpath);
			*/
		        result = -1;
			errno = EPERM;
		} else if(sscanf(line,"symlink %s %s",path,newpath)==2) {
			/*
			  temporarily disabling links because of a big security hole!
			  if(!chirp_path_fix(path)) goto failure;
			  if(!chirp_path_fix(newpath)) goto failure;
			  if(!chirp_acl_check(newpath,subject,CHIRP_ACL_WRITE)) goto failure;
			  result = symlink(path,newpath);
			*/
		        result = -1;
			errno = EPERM;
		} else if(sscanf(line,"setacl %s %s %s",path,newsubject,newacl)==3) {
			if(!chirp_path_fix(path)) goto failure;
			if(
				chirp_acl_check(path,subject,CHIRP_ACL_ADMIN)  ||
				chirp_acl_check_dir(path,subject,CHIRP_ACL_ADMIN) 
				) {
				result = chirp_acl_set(path,newsubject,chirp_acl_text_to_flags(newacl));
			} else {
				goto failure;
			}
		} else if(sscanf(line,"mkdir %s %lld",path,&mode)==2) {
			if(!chirp_path_fix(path)) goto failure;
			if( chirp_acl_check_or( path, subject,
						CHIRP_ACL_RESERVE ) ) {
				result = mkdir(path,mode);
				if(result==0){
					if(chirp_acl_init_reserve(path,subject)) {
						result = 0;
					} else {
						rmdir(path);
						errno = EACCES;
						result = -1;
					}
				}
			} else if(chirp_acl_check(path,subject,CHIRP_ACL_WRITE)){
				result = mkdir(path,mode);
				if(result==0){
					if(chirp_acl_init_copy(path)) {
						result = 0;
					} else {
						rmdir(path);
						errno = EACCES;
						result = -1;
					}
				}
			} else {
				goto failure;
			}
		} else if(sscanf(line,"rmdir %s",path)==1) {
			if(!chirp_path_fix(path)) goto failure;
			if( chirp_acl_check(path,subject,CHIRP_ACL_WRITE) ||
			    chirp_acl_check(path,subject,CHIRP_ACL_DELETE) ||
			    chirp_acl_check_dir(path,subject,CHIRP_ACL_WRITE)||
			    chirp_acl_check_dir(path,subject,CHIRP_ACL_DELETE)
				) {
				result = chirp_acl_rmdir(path);
			} else {
				goto failure;
			}
		} else if(sscanf(line,"utime %s %lu %lu",path,&ut.actime,&ut.modtime)==3) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_WRITE)) goto failure;
			result = utime(path,&ut);
		} else if(sscanf(line,"fstat %lld",&fd)==1) {
			result = fstat64(fd,&statbuf);
			do_stat_result = 1;
		} else if(sscanf(line,"fstatfs %lld",&fd)==1) {
			result = fstatfs(fd,&statfsbuf);
			do_statfs_result = 1;
		} else if(sscanf(line,"statfs %s",path)==1) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_LIST)) goto failure;
			result = statfs(path,&statfsbuf);
			do_statfs_result = 1;
		} else if(sscanf(line,"stat %s",path)==1) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_LIST)) goto failure;
			result = stat64(path,&statbuf);
			do_stat_result = 1; 
		} else if(sscanf(line,"lstat %s",path)==1) {
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_LIST)) goto failure;
			result = lstat64(path,&statbuf);
			do_stat_result = 1;
		} else if(sscanf(line,"exec %s %s %s %[^\n]",infile,outfile,errfile,args)==4) {
			if(!allow_execute) {
				errno = EPERM;
				goto failure;
			}
			if(!strcmp(infile,"-")) {
				strcpy(infile,"/dev/null");
			} else {
				if(!chirp_path_fix(infile)) goto failure;
				if(!chirp_acl_check(infile,subject,CHIRP_ACL_READ)) goto failure;
			}
			if(!strcmp(outfile,"-")) {
				strcpy(outfile,"/dev/null");
			} else {
				if(!chirp_path_fix(outfile)) goto failure;
				if(!chirp_acl_check(outfile,subject,CHIRP_ACL_WRITE)) goto failure;
			}
			if(!strcmp(errfile,"-")) {
				strcpy(errfile,"/dev/null");
			} else {
				if(!chirp_path_fix(errfile)) goto failure;
				if(!chirp_acl_check(errfile,subject,CHIRP_ACL_WRITE)) goto failure;
			}
			sscanf(args,"%s",path);
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_EXECUTE)) goto failure;
			result = chirp_exec(l,subject,infile,outfile,errfile,path,args);
		} else if(sscanf(line,"md5 %s",path)==1) {
			dataout = xxmalloc(16);
			if(!chirp_path_fix(path)) goto failure;
			if(!chirp_acl_check(path,subject,CHIRP_ACL_READ)) goto failure;			
			if(md5_file(path,dataout)) {
				result = dataoutlength = 16;
			} else {
				result = errno_to_chirp(errno);
			}		
		} else {
			result = -1;
			errno = ENOSYS;
		}

		if(do_no_result) {
			/* nothing */
		} else if(result<0) {
		failure:
			result = errno_to_chirp(errno);
			sprintf(line,"%lld\n",result);
		} else if(do_stat_result) {
			sprintf(line,"%lld\n%lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld\n",
				(long long) result,
				(long long) statbuf.st_dev,
				(long long) statbuf.st_ino,
				(long long) statbuf.st_mode,
				(long long) statbuf.st_nlink,
				(long long) statbuf.st_uid,
				(long long) statbuf.st_gid,
				(long long) statbuf.st_rdev,
				(long long) statbuf.st_size,
				(long long) statbuf.st_blksize,
				(long long) statbuf.st_blocks,
				(long long) statbuf.st_atime,
				(long long) statbuf.st_mtime,
				(long long) statbuf.st_ctime);
		} else if(do_statfs_result) {
			sprintf(line,"%lld\n%lld %lld %lld %lld %lld %lld %lld\n",
				(long long) result,
#if CCTOOLS_OPSYS_SUNOS
				(long long) statfsbuf.f_fsid,
#else
				(long long) statfsbuf.f_type,
#endif
       				(long long) statfsbuf.f_bsize,
				(long long) statfsbuf.f_blocks,
				(long long) statfsbuf.f_bfree,
				(long long) statfsbuf.f_bavail,
				(long long) statfsbuf.f_files,
				(long long) statfsbuf.f_ffree);
		} else if(do_getdir_result) {
			sprintf(line,"\n");
		} else {
			sprintf(line,"%lld\n",result);
		}

		debug(D_CHIRP,"= %s",line);
		if(!do_no_result) {
			length = strlen(line);
			actual = link_write(l,line,length,stalltime);
			if(actual!=length) break;
		}

		if(dataout) {
			actual = link_write(l,dataout,dataoutlength,stalltime);
			if(actual!=dataoutlength) break;
			free(dataout);
			dataout = 0;
		}

	}
}

static int errno_to_chirp( int e )
{
	switch(e) {
	case EACCES:
	case EPERM:
		return CHIRP_ERROR_NOT_AUTHORIZED;
	case ENOENT:
		return CHIRP_ERROR_DOESNT_EXIST;
	case EEXIST:
		return CHIRP_ERROR_ALREADY_EXISTS;
	case EFBIG:
		return CHIRP_ERROR_TOO_BIG;
	case ENOSPC:
		return CHIRP_ERROR_NO_SPACE;
	case ENOMEM:
		return CHIRP_ERROR_NO_MEMORY;
	case ENOSYS:
	case EINVAL:
		return CHIRP_ERROR_INVALID_REQUEST;
	case EMFILE:
	case ENFILE:
		return CHIRP_ERROR_TOO_MANY_OPEN;
	case EBUSY:
		return CHIRP_ERROR_BUSY;
	case EAGAIN:
		return CHIRP_ERROR_TRY_AGAIN;
	case EBADF:
		return CHIRP_ERROR_BAD_FD;
	case EISDIR:
		return CHIRP_ERROR_IS_DIR;
	case ENOTDIR:
		return CHIRP_ERROR_NOT_DIR;
	case ENOTEMPTY:
		return CHIRP_ERROR_NOT_EMPTY;
	case EXDEV:
		return CHIRP_ERROR_CROSS_DEVICE_LINK;
	default:
		debug(D_CHIRP,"zoiks, I don't know how to transform error %d (%s)\n",errno,strerror(errno));
		return CHIRP_ERROR_UNKNOWN;
	}
}
