#include "quota.h"
#include "debug.h"
#include "stringtools.h"

#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/param.h>
#ifdef HAS_SYS_STATFS_H
#include <sys/statfs.h>
#endif
#ifdef HAS_SYS_STATVFS_H
#include <sys/statvfs.h>
#endif

/* Cygwin does not have 64-bit I/O, while Darwin has it by default. */

#if CCTOOLS_OPSYS_CYGWIN || CCTOOLS_OPSYS_DARWIN
#define fopen64 fopen
#define open64 open
#define lseek64 lseek
#define stat64 stat
#define fstat64 fstat
#define lstat64 lstat
#define fseeko64 fseeko
#endif

/* Solaris has statfs, but it doesn't work! Use statvfs instead. */
#if CCTOOLS_OPSYS_SUNOS
#define statfs statvfs
#define fstatfs fstatvfs
#endif

#if CCTOOLS_OPSYS_DARWIN
typedef int blkcnt_t;
typedef int blksize_t;
#endif

////// the various quota routines
static int get_fd_info( int fd, blkcnt_t *blocks, off_t *size )
{
	struct stat64 st;
	int ret;

	ret = fstat64( fd, &st );
	if( blocks )
		*blocks = st.st_blocks;
	if( size )
		*size = st.st_size;
	return ret;
}

static int get_path_info( const char *path, blkcnt_t *blocks, off_t *size )
{
	struct stat st;
	int ret;

	ret = stat( path, &st );
	if( blocks )
		*blocks = st.st_blocks;
	if( size )
		*size = st.st_size;
	return ret;
}

static int get_fsf_info( int fd, long *blksize )
{
	int ret;
	struct stat64 st;

	ret = fstat64( fd, &st );
	if( blksize )
		*blksize = st.st_blksize;

	return ret;
}

static int is_sparse( char *buffer, size_t size )
{
	size_t x;
	for( x = 0; x < size; ++x, ++buffer ) {
		if( *buffer != 0 )
			return 0;
	}

	return 1;
}

ssize_t quota_write( struct limit *l, int f, const void *b, size_t c )
{
	blkcnt_t oldblocks, newblocks, netblocks;
	off_t oldsize, newsize, offset, n_offset;
	ssize_t retval, written = 0;
	int check_blocksize;
	char *buffer;
	long blocksize, can_write;

	if( l ){
		debug( D_CHIRP, "using quota" );
		//get filesystem blocksize
		if( get_fsf_info( f, &blocksize ) == 0 ){
			check_blocksize = blocksize / 512 + ( blocksize % 512 ) ? ( 1 ) : ( 0 );
			//get the current file pointer
			offset = lseek64( f, 0, SEEK_CUR );
      
			if( offset != ( off_t ) -1 ){
				//figure out where we are in the block
				offset %= blocksize;
				can_write = blocksize - offset;
	
				//get file size and allocated blocks
				if( get_fd_info( f, &oldblocks, &oldsize ) == 0 ){
	  
					//keep going as long as we have blocks and stuff to write
					while( limit_size_check( l ) >= check_blocksize &&
					       ( c > 0 ) ){
						retval = write( f, b, ( c < can_write ) ? c : can_write );
						if( retval < 0 ) {
							//return what's written so far
							if( written > 0 )
								return written;
							else
								return -1;
						}
						if( get_fd_info( f, &newblocks, &newsize ) == 0 ){
							netblocks = newblocks - oldblocks;
							if( netblocks > check_blocksize ){
								//this means that the write caused an indirect block
								//to be created (at the end of the file)
								//so we need to check to see if there is enough space
								if( limit_size_check( l ) < netblocks ){
									//not enough space, so truncate
									ftruncate( f, oldsize );
									//return what written so far
									if( written > 0 )
										return written;
									else{
										//no space left
										errno = ENOSPC;
										return -1;
									}
								}
							}
							else {
								//handle get_fd_info failure
								debug( D_CHIRP, "cannot get inode information" );
								return -1;
							}
	      
							//update quota information
							limit_do( l, netblocks );
	      
							//make the new size old
							oldblocks = newblocks;
							oldsize = newsize;
	      
							//update amount written so far, write pointer, and amount left
							written += retval;
							b += retval;
							c -= retval;
							if( can_write - retval != 0 )
								can_write -= retval;
							else
								can_write = blocksize;
						}
						else {
							//return what written so far
							if( written > 0 )
								return written;
							else{
								//no space left
								errno = ENOSPC;
								return -1;
							}
						}
					}
					//check to see if we're done
					if( c > 0 ) {
						buffer = ( char * ) malloc( blocksize );
						if( buffer ){
							while( c > 0 ){
								//stuff left to write but no blocks
								//figure out where we are
								offset = lseek64( f, 0, SEEK_CUR );
								n_offset = offset / blocksize;
								n_offset *= blocksize;
								lseek64( f, n_offset, SEEK_SET );
		
								//pull entire block
								retval = read( f, buffer, blocksize );
		
								//if all zeros, die w/ "no more room" error
								if( is_sparse( buffer, blocksize ) ){
									//return written so far
									free( buffer );
								}
								//else, write to block boundary and check again
								else {
									lseek64( f, offset, SEEK_SET );
									can_write = offset % blocksize;
									can_write -= blocksize;
									retval = write( f, b, ( c < can_write ) ? c : can_write );
									if( retval > 0 ){
										written += retval;
										b += retval;
										c -= retval;
									}
									else {
										free( buffer );
										//return what written so far
										if( written > 0 )
											return written;
										else{
											//no space left
											errno = ENOSPC;
											return -1;
										}
									}
								}
							}
	      
							free( buffer );
							return written;
						}
						else{
							errno = ENOMEM;
							return -1;
						}
					}
					else
						//we are!
						return written;
				}
				else
					return -1;
			}
			else
				return -1;
		}
		else
			return -1;
	}
	else
		return write( f, b, c );
}

ssize_t quota_pwrite( struct limit *l, int f, const void *b, size_t c,
		      off_t o )
{
	ssize_t ret2;
	off_t ret1, current_offset = lseek64( f, 0, SEEK_CUR );

	if( l ){
		if( current_offset != ( off_t ) -1 ){
			ret1 = lseek64( f, o, SEEK_SET );
			if( ret1 != ( off_t ) -1 ) {
				ret2 = quota_write( l, f, b, c );
				if( ret2 >= 0 ){
					ret1 = lseek64( f, current_offset, SEEK_SET );
					if( ret1 != ( off_t ) -1 )
						return ret2;
					else
						return ( ssize_t ) -1;
				}
				else {
					lseek64( f, current_offset, SEEK_SET );
					return ( ssize_t ) -1;
				}
			}
			else
				return ( ssize_t ) -1;
		}
		else
			return ( ssize_t ) -1;
	}
	else
		return pwrite( f, b, c, o );
}

ssize_t quota_full_pwrite( struct limit *l, int fd, const void *buf,
			   size_t count, off_t offset )
{
        INT64_T total=0;
        INT64_T chunk=0;

        while( count>0 ) {
                chunk = quota_pwrite(l,fd,buf,count,offset);
                if(chunk<0) {
                        if(errno==EINTR) {
                                continue;
                        }else {
				break;
                        }
                } else if(chunk==0) {
                        break;
                } else {
                        total += chunk;
                        count -= chunk;
                        buf += chunk;
                        offset += chunk;
                }
        }

        if(total>0) {
                return total;
        } else {
                if(chunk==0) {
                        return 0;
                } else {
                        return -1;
                }
        }
}


int quota_truncate( struct limit *l, const char *p, off_t len )
{
	blkcnt_t oldblocks, newblocks, netblocks;
	off_t oldsize;
	int ret, temp_errno;
  
	if( l ){
		if( ( ret = get_path_info( p, &oldblocks, &oldsize ) ) == 0 &&
		    ( ret = truncate( p, len ) ) == 0 ){
			if( get_path_info( p, &newblocks, NULL ) == 0 ) {
				netblocks = newblocks - oldblocks;
				if( limit_size_check( l ) >= netblocks ) {
					limit_do( l, netblocks );
					return ret;
				}
				else {
					temp_errno = errno;
					truncate( p, oldsize );
					errno = temp_errno;
					return -1;
				}
			}
			else{
				temp_errno = errno;
				truncate( p, oldsize );
				errno = temp_errno;
				return -1;
			}
		}
		else
			return ret;
	}
	else
		return truncate( p, len );
}

int quota_ftruncate( struct limit *l, int f, off_t len )
{
	blkcnt_t oldblocks, newblocks, netblocks;
	off_t oldsize;
	int ret, temp_errno;
  
	if( l ){
		if( ( ret = get_fd_info( f, &oldblocks, &oldsize ) ) == 0 &&
		    ( ret = ftruncate( f, len ) ) == 0 ){
			if( get_fd_info( f, &newblocks, NULL ) == 0 ) {
				netblocks = newblocks - oldblocks;
				if( limit_size_check( l ) >= netblocks ) {
					limit_do( l, netblocks );
					return ret;
				}
				else {
					temp_errno = errno;
					ftruncate( f, oldsize );
					errno = temp_errno;
					return -1;
				}
			}
			else{
				temp_errno = errno;
				ftruncate( f, oldsize );
				errno = temp_errno;
				return -1;
			}
		}
		else
			return ret;
	}
	else
		return ftruncate( f, len );
}


int quota_link( struct limit *l, const char *o, const char *n )
{
	blkcnt_t oldblocks, newblocks, netblocks;
	int ret, temp_errno, n_len;
	char *temp, *parent;

	if( l ){
		//allocate space large enough for the temp and the parent
		n_len = strlen( n ) + 1;
		temp = ( char * ) malloc( n_len + 3 );
		parent = ( char * ) malloc( n_len );

		if( temp && parent ) {
			//find the parent directory of n
			sprintf( temp, "%s/..", n );
			string_collapse_path(temp, parent, 1);
			if( ( ret = get_path_info( parent, &oldblocks, NULL ) ) == 0 &&
			    ( ret = link( o, n ) ) == 0 ) {
				if( get_path_info( parent, &newblocks, NULL ) == 0 ) {
					netblocks = newblocks - oldblocks;
					if( limit_size_check( l ) >= 0 ) {
						limit_do( l, netblocks );
						return ret;
					}
					else {
						temp_errno = errno;
						unlink( n );
						errno = temp_errno;
						return -1;
					}
				}
				else{
					temp_errno = errno;
					unlink( n );
					errno = temp_errno;
					return -1;
				}
			}
			else
				return ret;
		}
		else{
			errno = ENOMEM;
			return -1;
		}
	}
	else
		return link( o, n );
}


int quota_mkdir( struct limit *l, const char *p, mode_t m )
{
	blkcnt_t oldblocks, newblocks, netblocks, otherblocks;
	int ret, temp_errno, p_len;
	char *temp, *parent;

	if( l ){
		//allocate space large enough for the temp and the parent
		p_len = strlen( p ) + 1;
		temp = ( char * ) malloc( p_len + 3 );
		parent = ( char * ) malloc( p_len );

		if( temp && parent ) {
			//find the parent directory of n
			sprintf( temp, "%s/..", p );
			string_collapse_path(temp, parent, 1);
			if( ( ret = get_path_info( parent, &oldblocks, NULL ) ) == 0 &&
			    ( ret = mkdir( p, m ) ) == 0 ) {
				if( get_path_info( parent, &newblocks, NULL ) == 0 &&
				    get_path_info( p, &otherblocks, NULL ) == 0 ) {
					netblocks = newblocks + otherblocks - oldblocks;
					if( limit_size_check( l ) >= netblocks ) {
						limit_do( l, netblocks );
						return ret;
					}
					else {
						temp_errno = errno;
						rmdir( p );
						errno = temp_errno;
						return -1;
					}
				}
				else{
					temp_errno = errno;
					rmdir( p );
					errno = temp_errno;
					return -1;
				}
			}
			else
				return ret;
		}
		else{
			errno = ENOMEM;
			return -1;
		}
	}
	else
		return mkdir( p, m );
}

int quota_rmdir( struct limit *l, const char *p )
{
	blkcnt_t oldblocks, newblocks, netblocks, otherblocks;
	int ret, p_len;
	char *temp, *parent;

	if( l ){
		//allocate space large enough for the temp and the parent
		p_len = strlen( p ) + 1;
		temp = ( char * ) malloc( p_len + 3 );
		parent = ( char * ) malloc( p_len );

		if( temp && parent ) {
			//find the parent directory of n
			sprintf( temp, "%s/..", p );
			string_collapse_path(temp, parent, 1);
			if( ( ret = get_path_info( parent, &oldblocks, NULL ) ) == 0 &&
			    ( ret = get_path_info( p, &otherblocks, NULL ) ) == 0 &&
			    ( ret = rmdir( p ) ) == 0 ) {
				if( get_path_info( parent, &newblocks, NULL ) ) {
					netblocks = newblocks - otherblocks - oldblocks;
					if( limit_size_check( l ) >= netblocks ) {
						limit_do( l, netblocks );
						return ret;
					}
					else {
						//temp_errno = errno;
						//rmdir( p );
						//errno = temp_errno;
						return -1;
					}
				}
				else{
					//temp_errno = errno;
					//rmdir( p );
					//errno = temp_errno;
					return -1;
				}
			}
			else
				return ret;
		}
		else{
			errno = ENOMEM;
			return -1;
		}
	}
	else
		return rmdir( p );
}

int quota_open( struct limit *lf, struct limit *ld, const char *p, int f,
		mode_t m )
{
	blkcnt_t olddirblocks, newdirblocks, netdirblocks, oldfileblocks,
		newfileblocks, netfileblocks;
	int ret, temp_errno, p_len;
	char *temp, *parent;

	//get the file information like normal
	if ( ( f & O_TRUNC ) && ( lf ) ) {
		if( get_path_info( p, &oldfileblocks, NULL ) != 0 )
			return -1;
	}

	//find the parent dir if we are going to attempt to create a file
	if( ( f & O_CREAT ) && ( ld ) ){
		p_len = strlen( p ) + 1;
		temp = ( char * ) malloc( p_len + 3 );
		parent = ( char * ) malloc( p_len );
		if( temp && parent ){
			//find the parent directory
			sprintf( temp, "%s/..", p );
			string_collapse_path( temp, parent, 1 );
			if( get_path_info( parent, &olddirblocks, NULL ) != 0 ){
				free( parent );
				free( temp );
				return -1;
			}
		}
		else {
			errno = ENOMEM;
			return -1;
		}
	}

	ret = open64( p, m, f );

	if( ( f & O_CREAT ) && ( ld ) ){
		if( get_path_info( parent, &newdirblocks, NULL ) != 0 ){
			temp_errno = errno;
			unlink( p );
			free( temp );
			free( parent );
			close( ret );
			errno = temp_errno;
			return -1;
		}
		netdirblocks = newdirblocks - olddirblocks;
		if( limit_size_check( ld ) < netdirblocks ) {
			unlink( p );
			free( temp );
			free( parent );
			close( ret );
			errno = ENOSPC;
			return -1;
		}
		free( temp );
		free( parent );
	}

	if ( ( f & O_TRUNC ) && ( lf ) ) {
		if( get_path_info( p, &newfileblocks, NULL ) != 0 ){
			temp_errno = errno;
			unlink( p );
			errno = temp_errno;
			close( ret );
			return -1;
		}
		netfileblocks = newfileblocks - oldfileblocks;
	}

	if( ld )
		limit_do( ld, netdirblocks );
	if( lf )
		limit_do( lf, netfileblocks );

	return ret;
}


int quota_unlink( struct limit *lf, struct limit *ld, const char *p )
{
	struct stat64 st;
	int ret, p_len;
	blkcnt_t dirblocks, newdirblocks, fileblocks;
	char *temp, *parent;
	
	if( lf != NULL || ld != NULL ){
		
		if( stat64( p, &st ) == 0 ){
			//check to see if we need to keep track of file blocks
			if( st.st_nlink == 1 && lf ){
				fileblocks = st.st_blocks;
			}
			
			if( ld ){
				p_len = strlen( p ) + 1;
				temp = ( char * ) malloc( p_len + 3 );
				parent = ( char * ) malloc( p_len );
				if( temp && parent ){
					//find the parent directory
					sprintf( temp, "%s/..", p );
					string_collapse_path( temp, parent, 1 );
					free( temp );
					if( get_path_info( parent, &dirblocks, NULL ) != 0 ){
						free( parent );
						return -1;
					}
				}
				else {
					errno = ENOMEM;
					return -1;
				}
			}
		}
		else
			return -1;
		
		ret = unlink( p );
		
		if( ret == 0 ) {
			if( lf )
				limit_do( lf, -fileblocks );
			if( ld ){
				if( get_path_info( parent, &newdirblocks, NULL ) == 0 )
					limit_do( ld, newdirblocks - dirblocks );
				free( parent );
			}
		}
		
		return ret;
		
	}
	else
		return unlink( p );
}
/////// end the various quota routines
