/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.
*/

#include "uid_cache.h"
#include "hash_cache.h"
#include "xmalloc.h"

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <stdio.h>

#define LIFETIME 3600
#define KEY_MAX USERNAME_MAX

static struct hash_cache *name_to_uid=0;
static struct hash_cache *uid_to_name=0;
static struct hash_cache *name_to_gid=0;
static struct hash_cache *gid_to_name=0;

static int uid_cache_init()
{
	if(!name_to_uid) {
		name_to_uid = hash_cache_create(127,hash_string,free);
		if(!name_to_uid) return 0;
	}
	if(!uid_to_name) {
		uid_to_name = hash_cache_create(127,hash_string,free);
		if(!uid_to_name) return 0;
	}
	if(!name_to_gid) {
		name_to_gid = hash_cache_create(127,hash_string,free);
		if(!name_to_gid) return 0;
	}
	if(!gid_to_name) {
		gid_to_name = hash_cache_create(127,hash_string,free);
		if(!gid_to_name) return 0;
	}
	return 1;
}

int uid_cache_lookup( uid_t uid, char *name )
{
	char key[KEY_MAX];
	char *result;
	struct passwd *p;

	if(!uid_cache_init()) return 0;

	sprintf(key,"%d",uid);

	result = hash_cache_lookup(uid_to_name,key);
	if(result) {
		strcpy(name,result);
		return 1;
	}

	p = getpwuid(uid);
	if(p) {
		strcpy(name,p->pw_name);
		hash_cache_insert(uid_to_name,key,xstrdup(name),LIFETIME);
		return 1;
	} else {
		return 0;
	}
}

int uid_cache_lookup_reverse( const char *name, uid_t *uid )
{
	char key[KEY_MAX];
	char *result;
	struct passwd *p;

	if(!uid_cache_init()) return 0;

	result = hash_cache_lookup(name_to_uid,name);
	if(result) {
		*uid = atoi(result);
		return 1;
	}

	p = getpwnam(name);
	if(p) {
		*uid = p->pw_uid;
		sprintf(key,"%d",*uid);
		hash_cache_insert(name_to_uid,name,xstrdup(key),LIFETIME);
		return 1;
	} else {
		return 0;
	}
}

int gid_cache_lookup( gid_t gid, char *name )
{
	char key[KEY_MAX];
	char *result;
	struct group *g;

	if(!uid_cache_init()) return 0;

	sprintf(key,"%d",gid);

	result = hash_cache_lookup(gid_to_name,key);
	if(result) {
		strcpy(name,result);
		return 1;
	}

	g = getgrgid(gid);
	if(g) {
		strcpy(name,g->gr_name);
		hash_cache_insert(gid_to_name,key,xstrdup(name),LIFETIME);
		return 1;
	} else {
		return 0;
	}
}

int gid_cache_lookup_reverse( const char *name, gid_t *gid )
{
	char key[KEY_MAX];
	char *result;
	struct group *g;

	if(!uid_cache_init()) return 0;

	result = hash_cache_lookup(name_to_gid,name);
	if(result) {
		*gid = atoi(result);
		return 1;
	}

	g = getgrnam(name);
	if(g) {
		*gid = g->gr_gid;
		sprintf(key,"%d",*gid);
		hash_cache_insert(name_to_gid,name,xstrdup(key),LIFETIME);
		return 1;
	} else {
		return 0;
	}
}

