/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.
*/

#include "pfs_service.h"

extern "C" {
#include "debug.h"
#include "stringtools.h"
#include "domain_name.h"
#include "link.h"
#include "file_cache.h"
#include "full_io.h"
#include "http_query.h"
#include "hash_table.h"
#include "xmalloc.h"
}

#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/statfs.h>

#define HTTPFS_LINE_MAX 4096
#define HTTPFS_PORT 80

#define HTTP_TYPE_DIR	0
#define HTTP_TYPE_FILE	1
#define HTTP_TYPE_LINK	2

#define HTTPFS_FILE_MODE (S_IFREG | 0555)
#define HTTPFS_DIR_MODE (S_IFDIR | 0555)
#define HTTPFS_LINK_MODE (S_IFLNK | 0555)

extern int pfs_master_timeout;

/*
Keep a hash table mapping full pathnames to already-built
stat structures.  This filesystem never changes, so it's
ok to keep them indefinitely.
*/

static struct hash_table *stat_cache=0;

/*
Fetch an item and determine whether it is a link, a directory, or a file.
A link with have a .parrothttpfslink entry describing the link target.
A directory will have a .parrothttpfsdir entry containing its contents,
while a file is just a file.  Try the link, then the dir, then the file,
and indicate in "type" which was successful.
*/

struct link * httpfs_fetch( pfs_name *name, const char *action, int *type, int *size, int want_link )
{
	char url[HTTPFS_LINE_MAX];
	struct link *result;
	*size = 0;

	if(!name) abort();
	if(!name->host[0]) {
		errno = ENOENT;
		return 0;
	}

	if(want_link) {
		char dirname[PFS_PATH_MAX];
		const char *basename;
		int length;

		basename = string_basename(name->rest);
		length = basename-name->rest;
		strncpy(dirname,name->rest,length);
		dirname[length] = 0;
		debug(D_HTTP,"dirname: %s  basename: %s",dirname,basename);

		sprintf(url,"http://%s:%d%s/.parrothttpfslink-%s",name->host,name->port,dirname,basename);
		result = http_query_size(url,action,size,time(0)+pfs_master_timeout);
		if(result) {
			*type = HTTP_TYPE_LINK;
			return result;
		}
	}

	sprintf(url,"http://%s:%d%s/.parrothttpfsdir",name->host,name->port,name->rest);
	result = http_query_size(url,action,size,time(0)+pfs_master_timeout);
	if(result) {
		*type = HTTP_TYPE_DIR;
		return result;
	}

	sprintf(url,"http://%s:%d%s",name->host,name->port,name->rest);
	result = http_query_size(url,action,size,time(0)+pfs_master_timeout);
	if(result) {
		*type = HTTP_TYPE_FILE;
		return result;
	}

	return 0;
}

class pfs_file_httpfs : public pfs_file
{
private:
	struct link *link;
	int size;

public:
	pfs_file_httpfs( pfs_name *n, struct link *l, int s ) : pfs_file(n) {
		link = l;
		size = s;
	}

	virtual int close() {
		link_close(link);
		return 0;
	}

	virtual pfs_ssize_t read( void *d, pfs_size_t length, pfs_off_t offset ) {
		return link_read(link,(char*)d,length,LINK_FOREVER);
	}

	virtual int fstat( struct pfs_stat *buf ) {
		pfs_service_emulate_stat(&name,buf);
		buf->st_mode = HTTPFS_FILE_MODE;
		buf->st_size = size;
		return 0;
	}

	virtual pfs_ssize_t get_size() {
		return size;
	}

};

class pfs_service_httpfs : public pfs_service {
public:
	virtual int get_default_port() {
		return HTTPFS_PORT;
	}

	virtual pfs_file * open( pfs_name *name, int flags, mode_t mode ) {
		struct link *link;
		int type, size;

		if((flags&O_ACCMODE)!=O_RDONLY) {
			errno = EROFS;
			return 0;
		}
 
		link = httpfs_fetch(name,"GET",&type,&size,0);
		if(link) {
			if(type==HTTP_TYPE_DIR) {
				link_close(link);
				errno = EISDIR;
				return 0;
			} else {
				return new pfs_file_httpfs(name,link,size);
			}
		} else {
			return 0;
		}
	}

	pfs_dir * getdir( pfs_name *name ) {
		char line[PFS_PATH_MAX];
		struct link *link;
		int type,size;

		link = httpfs_fetch(name,"GET",&type,&size,0);
		if(link) {
			if(type==HTTP_TYPE_DIR) {
				pfs_dir *dir = new pfs_dir(name);

				while(link_readline(link,line,sizeof(line),time(0)+pfs_master_timeout)) {
					dir->append(line);
				}
				link_close(link);
				return dir;
			} else {
				link_close(link);
				errno = ENOTDIR;
				return 0;
			}
		} else {
			return 0;
		}
	}

	virtual int lstat( pfs_name *name, struct pfs_stat *buf ) {
		char tag[PFS_PATH_MAX];
		struct pfs_stat *cacheresult;
		struct link *link;
		int type, size;

		sprintf(tag,"http://%s:%d%s",name->host,name->port,name->rest);

		if(!stat_cache) stat_cache = hash_table_create(0,0);

		cacheresult = (struct pfs_stat *) hash_table_lookup(stat_cache,tag);
		if(cacheresult) {
			debug(D_HTTP,"using cached stat() of %s",tag);
			if(cacheresult->st_dev>0) {
				errno = cacheresult->st_dev;
				return -1;
			} else {
				memcpy(buf,cacheresult,sizeof(*buf));
				return 0;
			}
		}

		cacheresult = new pfs_stat;

		link = httpfs_fetch(name,"HEAD",&type,&size,1);
		if(link) {
			link_close(link);
			pfs_service_emulate_stat(name,buf);
			if(type==HTTP_TYPE_DIR) {
				buf->st_mode = HTTPFS_DIR_MODE;
			} else if(type==HTTP_TYPE_FILE) {
				buf->st_size = size;
				buf->st_mode = HTTPFS_FILE_MODE;
			} else if(type==HTTP_TYPE_LINK) {
				buf->st_size = size;
				buf->st_mode = HTTPFS_LINK_MODE;
			}
			buf->st_dev = -1;
			memcpy(cacheresult,buf,sizeof(*buf));
			hash_table_insert(stat_cache,tag,cacheresult,(void**)0);
			return 0;
		} else {
			cacheresult->st_dev = errno;
			hash_table_insert(stat_cache,tag,cacheresult,(void**)0);
			return -1;
		}	
	}

	virtual int stat( pfs_name *name, struct pfs_stat *buf ) {
		int result;
		result = lstat(name,buf);
		if(result==0) {
			if(buf->st_mode==HTTPFS_LINK_MODE) {
				buf->st_mode = HTTPFS_DIR_MODE;
				buf->st_size = 0;
			}
		}
		return result;
	}

	virtual int unlink( pfs_name *name ) {
		errno = EROFS;
		return -1;
	}

	virtual int access( pfs_name *name, mode_t mode ) {
		int result;
		struct pfs_stat buf;

		result = this->stat(name,&buf);
		if(result==0) {
			if(mode&W_OK) {
				errno = EROFS;
				return -1;
			} else {
				return 0;
			}
		} else {
			/* errno set by stat */
			return -1;
		}
	}

	virtual int chmod( pfs_name *name, mode_t mode ) {
		errno = EROFS;
		return -1;
	}

	virtual int chown( pfs_name *name, uid_t uid, gid_t gid ) {
		errno = EROFS;
		return -1;
	}

	virtual int lchown( pfs_name *name, uid_t uid, gid_t gid ) {
		errno = EROFS;
		return -1;
	}

	virtual int truncate( pfs_name *name, pfs_off_t length ) {
		errno = EROFS;
		return -1;
	}

	virtual int utime( pfs_name *name, struct utimbuf *buf ) {
		errno = EROFS;
		return -1;
	}

	virtual int rename( pfs_name *oldname, pfs_name *newname ) {
		errno = EROFS;
		return -1;
	}

	virtual int chdir( pfs_name *name, char *newpath ) {
		int result;
		struct pfs_stat buf;

		result = stat(name,&buf);
		if(result==0) {
			if(S_ISDIR(buf.st_mode)) {
				return 0;
			} else {
				errno = ENOTDIR;
				return -1;
			}
		} else {
			/* errno set by stat */
			return -1;
		}
	}

	virtual int link( pfs_name *oldname, pfs_name *newname ) {
		errno = EROFS;
		return -1;
	}

	virtual int symlink( pfs_name *oldname, pfs_name *newname ) {
		errno = EROFS;
		return -1;
	}

	virtual int readlink( pfs_name *name, char *buf, pfs_size_t bufsiz ) {
		int result;
		struct pfs_stat info;

		result = lstat(name,&info);
		if(result==0) {
			if(S_ISLNK(info.st_mode)) {
				struct link *link;
				int type, size;
				link = httpfs_fetch(name,"GET",&type,&size,1);
				if(link) {
					int length = link_read(link,buf,bufsiz,time(0)+pfs_master_timeout);
					link_close(link);
					// note that the last char is a newline
					buf[length] = 0;
					return length-1;
				}
				// errno set by fetch
				return -1;
			} else {
				errno = EINVAL;
				return -1;
			}
		} else {
			// errno set by lstat
			return -1;
		}
	}

	virtual int mkdir( pfs_name *name, mode_t mode ) {
		errno = EROFS;
		return -1;
	}

	virtual int rmdir( pfs_name *name ) {
		errno = EROFS;
		return -1;
	}

};

static pfs_service_httpfs pfs_service_httpfs_instance;
pfs_service *pfs_service_httpfs = &pfs_service_httpfs_instance;


