/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.
*/

#ifndef PFS_SYSDEPS_H
#define PFS_SYSDEPS_H

/*
In this file, we establish all of the definitions
and feature activations that are dependent upon
each system, with some explanation as to why they
are necessary.

Notice that we simply define our own private
versions of structures like stat, stat64, and so forth.
These are *not* the user level versions of these
structure, *nor* the kernel level version.  These are
the structures used at the kernel interface, which
are occasionally different than the other two, and
remarkably difficult to pull in a definition from
the right include files.  So, we just define our own.
*/

#ifndef _LARGEFILE64_SOURCE
#define _LARGEFILE64_SOURCE
#endif

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include "pfs_types.h"

#if defined(linux)

#include <sched.h>
#include <linux/net.h>

/*
On Linux,the arguments to mmap come indirectly
through a structure pointed to by the first arg.
*/

#define PFS_INDIRECT_MMAP_ARGS 1

/*
And, with mmap2, the offset in the file is given
in integral pages, not bytes
*/

#define PFS_MMAP_OFFSET_IN_PAGES 1

/*
On Linux, ptrace forces debugged processes to become
children of those that debug them.  Thus, we must keep
track of the process trees and deliver process completion
signals ourself.
*/

#define PFS_MANAGE_PROCESS_TREES 1

/*
On Linux, fork() does not cause the child process to
inherit the ptrace attribute.  Instead, we must
convert fork() into an equivalent clone() with the
ptrace attribute.  Further, the necessary flags
are not always defined in all kernels.
*/

#define PFS_CONVERT_FORK_TO_CLONE 1

#ifndef CLONE_PTRACE
#define CLONE_PTRACE    0x00002000
#endif

#ifndef CLONE_PARENT
#define CLONE_PARENT    0x00008000
#endif

/*
We require a scratch area in each process' address
space for storing internal data such as an adjusted
argument to execve.  This need not be *free* space,
as we will restore its contents afterwards.  It must
simply be a valid memory address.
*/

#define PFS_SCRATCH_ADDR ((char*)0x08048000)
#define PFS_SCRATCH_SIZE 4096

/*
Sadly, it is necessary to prefix each of the structure
tags with pfs (i.e. pfs_st_dev instead of st_dev) because
the structure tags themselves are frequently the target
of a macro definition that munges something at the user
level that does not concern us.
*/

struct pfs_kernel_stat {
	unsigned long st_dev;
	unsigned long st_ino;
	unsigned short st_mode;
	unsigned short st_nlink;
	unsigned short st_uid;
	unsigned short st_gid;
	unsigned long st_rdev;
	unsigned long st_size;
	unsigned long st_blksize;
	unsigned long st_blocks;
	struct timespec st_atim;
	struct timespec st_mtim;
	struct timespec st_ctim;
  /*	unsigned long st_atime;
	unsigned long st_p1;
	unsigned long st_mtime;
	unsigned long st_p2;
	unsigned long st_ctime;
	unsigned long st_p3;
  */
	unsigned long st_p4;
	unsigned long st_p5;
};

struct pfs_kernel_stat64 {
	unsigned long long st_dev;
	unsigned long st_pad1;
	unsigned long st_ino;
	unsigned long st_mode;
	unsigned long st_nlink;
	unsigned long st_uid;
	unsigned long st_gid;
	unsigned long long st_rdev;
	unsigned long st_pad2;
	unsigned long long st_size;
	unsigned long st_blksize;
	unsigned long long st_blocks;
	struct timespec st_atim;
	struct timespec st_mtim;
	struct timespec st_ctim;
/*
	unsigned long st_atime;
	unsigned long st_unused1;
	unsigned long st_mtime;
	unsigned long st_unused2;
	unsigned long st_ctime;
	unsigned long st_unused3;
*/
	unsigned long long st_ino_extra;
};

struct pfs_kernel_statfs {
        long f_type;
        long f_bsize;
        long f_blocks;
        long f_bfree;
        long f_bavail;
        long f_files;
        long f_ffree;
	fsid_t f_fsid;
        long f_namelen;
        long f_spare[6];
};

/* Linux has no statfs64 at the kernel interface yet! */

struct pfs_kernel_dirent {
        unsigned long   d_ino;
        unsigned long   d_off;
        unsigned short  d_reclen;
	unsigned char   d_type;
        char            d_name[PFS_PATH_MAX];
};

struct pfs_kernel_dirent64 {
        unsigned long long d_ino;
        unsigned long long d_off;
        unsigned short     d_reclen;
	unsigned char      d_type;
        char               d_name[PFS_PATH_MAX];
};

/*
Note that the typical libc sigaction places the
sa_mask field as the second value. This is hard to expand,
as the number of signals tends to increase.  The kernel
defined sigaction puts sa_mask last as follows:
*/

struct pfs_kernel_sigaction {
	void (*sa_handler)(int);
        long sa_flags;
        void (*sa_restorer)(void);
        sigset_t sa_mask;
};

#define PFS_GETLK	5
#define PFS_SETLK	6
#define PFS_SETLKW	7

#define PFS_GETLK64	12
#define PFS_SETLK64	13
#define PFS_SETLKW64	14

#else
#error This platform is not supported yet!
#endif

/*
Many data structures must be aligned on 8 byte boundaries.
This rounds up values to multiples of 8.
*/

#ifndef _ROUND_UP
#define _ROUND_UP(x,n) (((x)+(n)-1u) & ~((n)-1u))
#endif

#ifndef ROUND_UP
#define ROUND_UP(x) _ROUND_UP(x,8LL)
#endif

/*
The size of a dirent is the size of the structure
without the name field, plus the actual length
of the null-terminated name, rounded up to 8-byte alignment.
*/

#define DIRENT_SIZE( x ) \
	ROUND_UP(((char*)&(x).d_name[0]-(char*)&(x)) + strlen((x).d_name) + 1)

#define COPY_DIRENT( a, b ) \
	memset(&(b),0,sizeof((b))); \
	strcpy((b).d_name,(a).d_name);\
	(b).d_ino = (a).d_ino;\
	(b).d_off = (a).d_off;\
	(b).d_reclen = DIRENT_SIZE(b);

#endif
